using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using System.Xml;

using Smdn.Formats.Notations;
using Smdn.Formats.Notations.Dom;

namespace Smdn.Formats.Notations.Pakina {
  internal class pakina {
    private static Dictionary<string, Notation> availableNotations = new Dictionary<string, Notation>(StringComparer.InvariantCultureIgnoreCase);
    private static Dictionary<string, Encoding> unicodeEncodings = new Dictionary<string, Encoding>(StringComparer.InvariantCultureIgnoreCase) {
      {"utf8",        new UTF8Encoding(false, true)},
      {"utf8-bom",    new UTF8Encoding(true, true)},
      {"utf16le",     new UnicodeEncoding(false, false, true)},
      {"utf16be",     new UnicodeEncoding(true, false, true)},
      {"utf16le-bom", new UnicodeEncoding(false, true, true)},
      {"utf16be-bom", new UnicodeEncoding(true, true, true)},
      {"utf32le",     new UTF32Encoding(false, false, true)},
      {"utf32be",     new UTF32Encoding(true, false, true)},
      {"utf32le-bom", new UTF32Encoding(false, true, true)},
      {"utf32be-bom", new UTF32Encoding(true, true, true)},
    };
    private static Dictionary<string, string> unicodeEncodingNames = new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase) {
      {"utf8",        "UTF-8"},
      {"utf8-bom",    "UTF-8 (BOM)"},
      {"utf16le",     "UTF-16LE"},
      {"utf16be",     "UTF-16BE"},
      {"utf16le-bom", "UTF-16LE (BOM)"},
      {"utf16be-bom", "UTF-16BE (BOM)"},
      {"utf32le",     "UTF-32LE"},
      {"utf32be",     "UTF-32BE"},
      {"utf32le-bom", "UTF-32LE (BOM)"},
      {"utf32be-bom", "UTF-32BE (BOM)"},
    };

    private class Configuration {
      public string Input = null;
      public string Output = null;
      public string InputNotation = "pukiwiki";
      public string OutputNotation = null;
      public Encoding InputEncoding = Encoding.Default;
      public Encoding OutputEncoding = Encoding.UTF8;
      public bool Indent = true;
      public bool Fragment = false;
      public int IndentLevel = 2;
      public char IndentChar = ' ';
      public string Newline = Environment.NewLine;
      public string Title = null;
      public string DocumentElement = "body";
      public string DocumentElementPrefix = string.Empty;
      public string DocumentElementNamespaceUri = Smdn.Formats.Notations.XhtmlFormatter.XhtmlNamespaceUri;
      public bool PreservePreWhitespaces = false;
    }

    private static Configuration config = new Configuration();

    public static void Main(string[] args)
    {
      string doList = null;

      if (args.Length == 0)
        Usage(null);

      for (var index = 0; index < args.Length; index++) {
        switch (args[index]) {
          case "-h":
          case "--help":
            Usage(null);
            break;

          case "--version":
            Version();
            break;

          case "--list": doList = "notations"; break;
          case "--list-notations": doList = "notations"; break;
          case "--list-charsets": doList = "encodings"; break;

          case "-i":
            config.Input = args[++index];
            if (config.Input == "-")
              config.Input = null;
            break;

          case "-o":
            config.Output = args[++index];
            if (config.Output == "-")
              config.Output = null;
            break;

          case "-if": config.InputNotation = args[++index]; break;
          case "-of": config.OutputNotation = args[++index]; break;
          case "-ie": config.InputEncoding = GetEncoding(args[++index]); break;
          case "-oe": config.OutputEncoding = GetEncoding(args[++index]); break;
          case "--no-indent": config.Indent = false; break;
          case "--fragment": config.Fragment = true; break;
          case "--preserve-pre-whitespaces": config.PreservePreWhitespaces = true; break;

          default: {
            var delim = args[index].IndexOf("=");
            var arg = (0 <= delim) ? args[index].Substring(0, delim) : args[index];
            var val = (0 <= delim) ? args[index].Substring(delim + 1) : null;

            switch (arg) {
              case "--title": config.Title = val ?? string.Empty; break;
              case "--doc-element": config.DocumentElement = val ?? string.Empty; break;
              case "--doc-prefix": config.DocumentElementPrefix = val ?? string.Empty; break;
              case "--doc-namespace-uri": config.DocumentElementNamespaceUri = val ?? string.Empty; break;

              case "--indent-spaces":
                config.Indent = true;
                config.IndentChar = ' ';
                if (val != null)
                  config.IndentLevel = int.Parse(val);
                break;

              case "--indent-tabs":
                config.Indent = true;
                config.IndentChar = '\t';
                if (val != null)
                  config.IndentLevel = int.Parse(val);
                break;

              case "--newline":
                config.Indent = true;
                if (val != null) {
                  switch (val.ToLowerInvariant()) {
                    case "cr": config.Newline = "\u000d"; break;
                    case "lf": config.Newline = "\u000a"; break;
                    case "crlf": config.Newline = "\u000d\u000a"; break;
                    default: Usage("'{0}' is invalid newline chars", val); break;
                  }
                }
                break;
              case "":
                break;
            }

            break;
          } // default:
        } // switch
      } // for

      LookupNotations();

      if (doList == "notations")
        ListNotations();
      else if (doList == "encodings")
        ListEncodings();

      if (!availableNotations.ContainsKey(config.InputNotation))
        Usage("input notation '{0}' is not available", config.InputNotation);

      if (string.IsNullOrEmpty(config.OutputNotation))
        config.OutputNotation = config.InputNotation;
      else if (!availableNotations.ContainsKey(config.OutputNotation))
        Usage("output notation '{0}' is not available", config.OutputNotation);

      var parser = availableNotations[config.InputNotation].CreateParser();
      var formatter = availableNotations[config.OutputNotation].CreateXhtmlFormatter();

      var writerSettings = new XmlWriterSettings();

      writerSettings.ConformanceLevel = ConformanceLevel.Fragment;
      writerSettings.Indent = config.Indent;
      writerSettings.IndentChars = new string(config.IndentChar, config.IndentLevel);
      writerSettings.NewLineChars = config.Newline;
      writerSettings.NewLineOnAttributes = false;
      writerSettings.OmitXmlDeclaration = config.Fragment;
      writerSettings.CloseOutput = false;

      using (TextReader reader = (config.Input == null)
             ? new StreamReader(Console.OpenStandardInput(), config.InputEncoding)
             : new StreamReader(config.Input, config.InputEncoding)) {
        using (TextWriter writer = (config.Output == null)
               ? new StreamWriter(Console.OpenStandardOutput(), config.OutputEncoding)
               : new StreamWriter(config.Output, false, config.OutputEncoding)) {
          writer.NewLine = config.Newline;

          var continueParse = true;

          parser.FormFeed += delegate {
            continueParse = true;
          };

          while (continueParse) {
            continueParse = false;

            // parse
            var document = parser.Parse(reader);

            // format
            var xml = config.Fragment
              ? formatter.Format(document,
                                 config.DocumentElement,
                                 config.DocumentElementPrefix,
                                 config.DocumentElementNamespaceUri)
              : formatter.Format(document);

            if (!config.Fragment) {
              var body = xml.DocumentElement;

              xml.RemoveChild(body);

              xml.AppendChild(xml.CreateXmlDeclaration("1.0", config.OutputEncoding.WebName, null));

              var html = xml.AppendChild(xml.CreateElement("html", XhtmlFormatter.XhtmlNamespaceUri));
              var head = html.AppendChild(xml.CreateElement("head", XhtmlFormatter.XhtmlNamespaceUri));
              var title = head.AppendChild(xml.CreateElement("title", XhtmlFormatter.XhtmlNamespaceUri));

              body.RemoveAttribute("xmlns");

              html.AppendChild(body);

              if (config.Title != null || config.Input == null)
                title.AppendChild(xml.CreateTextNode(config.Title));
              else
                title.AppendChild(xml.CreateTextNode(config.Input));
            }

            var nsmgr = new XmlNamespaceManager(xml.NameTable);
            var preExists = false;

            nsmgr.AddNamespace("x", XhtmlFormatter.XhtmlNamespaceUri);

            foreach (XmlNode pre in xml.SelectNodes("//x:pre", nsmgr)) {
              preExists = true;

              if (config.PreservePreWhitespaces)
                pre.Attributes.Append(xml.CreateAttribute("xml:space")).Value = "preserve";

              // HACK: insert 'ZERO WIDTH SPACE' (U+200B) to avoid indenting
              pre.PrependChild(xml.CreateTextNode("\u200b"));
            }

            // output
            var sb = new StringBuilder();
            var w = XmlWriter.Create(sb, writerSettings);

            xml.WriteContentTo(w);

            w.Flush();

            if (preExists)
              writer.Write(sb.Replace("<pre>\u200b", "<pre>").ToString());
            else
              writer.Write(sb.ToString());

            if (continueParse)
              writer.WriteLine("\u000c");
            else
              writer.WriteLine();

            writer.Flush();
          } // while
        } // using writer
      } // using reader
    }

    private static void LookupNotations()
    {
      var notationType = typeof(Notation);
      var codeBaseDir = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
      var directories = (Environment.CurrentDirectory == codeBaseDir)
        ? new[] {Environment.CurrentDirectory}
        : new[] {Environment.CurrentDirectory, codeBaseDir};

      foreach (var directory in directories) {
        foreach (var file in Directory.GetFiles(directory, "*.dll", SearchOption.TopDirectoryOnly)) {
          try {
            Assembly.LoadFile(file);
          }
          catch (Exception ex) {
            Console.Error.WriteLine("assembly {0} is not loaded: {1}", file, ex);
          }
        }
      }

      foreach (var assm in AppDomain.CurrentDomain.GetAssemblies()) {
        foreach (var type in assm.GetExportedTypes()) {
          if (!(type.IsPublic && type.IsSubclassOf(notationType)))
            continue;

          var notation = (Notation)Activator.CreateInstance(type);

          if (availableNotations.ContainsKey(notation.Name)) {
            if (type.IsSubclassOf(availableNotations[notation.Name].GetType()))
              availableNotations.Remove(notation.Name);
            else
              continue;
          }

          availableNotations.Add(notation.Name, notation);
        }
      }
    }

    private static void Version()
    {
      var executingAssembly = Assembly.GetExecutingAssembly();
      var executingAssemblyName = executingAssembly.GetName();

      Console.WriteLine("{0} {1}.{2}", executingAssemblyName.Name, executingAssemblyName.Version.Major, executingAssemblyName.Version.Minor);

      Environment.Exit(0);
    }

    private static void ListNotations()
    {
      Console.WriteLine("available notations:");

      foreach (var notation in availableNotations) {
        Console.WriteLine("  {0,-15} {1} ({2})", notation.Key, notation.Value.GetType(), notation.Value.GetType().Assembly);
      }

      Console.WriteLine("total {0} notations", availableNotations.Count);

      Environment.Exit(0);
    }

    private static Encoding GetEncoding(string encoding)
    {
      if (unicodeEncodings.ContainsKey(encoding))
        return unicodeEncodings[encoding];
      else
        return Encoding.GetEncoding(encoding);
    }

    private static void ListEncodings()
    {
      Console.WriteLine("available encodings:");

      Console.WriteLine("  {0,-20} {1} (default)", Encoding.Default.WebName, Encoding.Default.EncodingName);

      foreach (var pair in unicodeEncodingNames) {
        Console.WriteLine("  {0,-20} {1}", pair.Key, pair.Value);
      }

      foreach (var enc in Encoding.GetEncodings()) {
        var encoding = enc.GetEncoding();

        Console.WriteLine("  {0,-20} {1}", encoding.WebName, encoding.EncodingName);
      }

      Environment.Exit(0);
    }

    private static void Usage(string format, params object[] args)
    {
      var defaults = new Configuration();

      Console.Error.WriteLine("usage:");
      Console.Error.WriteLine("  {0} [options]", Path.GetFileName(System.Reflection.Assembly.GetEntryAssembly().Location));
      Console.Error.WriteLine();
      Console.Error.WriteLine("general options:");
      Console.Error.WriteLine("  -h, --help       : show this message");
      Console.Error.WriteLine("  --list,");
      Console.Error.WriteLine("  --list-notations : list available notations");
      Console.Error.WriteLine("  --list-charsets  : list available encodings");
      Console.Error.WriteLine("  --version        : show version");
      Console.Error.WriteLine();
      Console.Error.WriteLine("  -i <file>        : input file");
      Console.Error.WriteLine("                       stdin will be used if file is '-' or omitted.");
      Console.Error.WriteLine("  -o <file>        : output file");
      Console.Error.WriteLine("                       stdout will be used if file is '-' or omitted.");
      Console.Error.WriteLine("  -if <notation>   : specify input notation, default='{0}'", defaults.InputNotation);
      Console.Error.WriteLine("  -of <notation>   : specify output notation, default=(same as input notation)");
      Console.Error.WriteLine("  -ie <charset>    : specify input encoding, default='{0}'", defaults.InputEncoding.WebName);
      Console.Error.WriteLine("  -oe <charset>    : specify output encoding, default='{0}'", defaults.OutputEncoding.WebName);
      Console.Error.WriteLine();
      Console.Error.WriteLine("output options:");
      Console.Error.WriteLine("  --no-indent                 : no indent");
      Console.Error.WriteLine("  --indent-spaces[=<number>]  : specify indent chars to <number>s of space");
      Console.Error.WriteLine("                                  default <number>={0}", defaults.IndentLevel);
      Console.Error.WriteLine("  --indent-tabs[=<number>]    : specify indent chars to <number>s of tab");
      Console.Error.WriteLine("                                  default <number>={0}", defaults.IndentLevel);
      Console.Error.WriteLine("  --newline[=(CR|LF|CRLF)]    : specify newline");
      Console.Error.WriteLine("                                  default=(same as system default)");
      Console.Error.WriteLine("  --title=<title>             : specify value of <title> element");
      Console.Error.WriteLine("                                  default=(input file name)");
      Console.Error.WriteLine("  --fragment                  : fragment mode");
      Console.Error.WriteLine("                                  omit <html> element, <head> element and xml declaration.");
      Console.Error.WriteLine("  --doc-element=<name>        : specify document element name");
      Console.Error.WriteLine("                                  default='{0}'", defaults.DocumentElement);
      Console.Error.WriteLine("                                  used if fragment mode.");
      Console.Error.WriteLine("  --doc-namespace-uri=<uri>   : specify document element namespace URI");
      Console.Error.WriteLine("                                  default='{0}'", defaults.DocumentElementNamespaceUri);
      Console.Error.WriteLine("                                  used if fragment mode.");
      Console.Error.WriteLine("  --doc-prefix=<prefix>       : specify document element namespace prefix");
      Console.Error.WriteLine("                                  default='{0}'", defaults.DocumentElementPrefix);
      Console.Error.WriteLine("                                  used if fragment mode.");
      Console.Error.WriteLine("  --preserve-pre-whitespaces  : append 'xml:space=\"preserve\"' to <pre> element");

      if (format == null) {
        Environment.Exit(0);
      }
      else {
        Console.Error.WriteLine();
        Console.Error.WriteLine("error:");
        Console.Error.Write("  ");
        Console.Error.WriteLine(format, args);
        Console.Error.WriteLine();

        Environment.Exit(-1);
      }
    }
  }
}
