// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Collections.Generic;
using System.Text.RegularExpressions;

using Smdn.Formats.Notations;
using Smdn.Formats.Notations.Dom;
using Smdn.Formats.Notations.PukiWiki.Dom;

namespace Smdn.Formats.Notations.PukiWiki {
  public class Parser : WikiLikeNotationParserBase {
    private static List<string> inlinePluginNames = new List<string>() {
      "t",
      "br",
      "counter",
      "online",
      "version",
      "page",
      "fpage",
      "date",
      "_date",
      "time",
      "_time",
      "now",
      "_now",
      "lastmod",
      "heart",
      // face
      "smile",
      "bigsmile",
      "huh",
      "oh",
      "wink",
      "sad",
      "worried",
    };

    private const string inlinePluginRegexString = @"\&(?<name>[\w\d_]+)(\((?<args>.*?)\))?";
    private Regex blockPluginRegex = new Regex(@"^#[a-zA-Z0-9]+(\([^\)]*\))?");
    private Regex headerAnchorRegex = new Regex(@"\[\#(?<id>[0-9a-z]{8})\]$");

    public bool ConvertLineBreak {
      get; set;
    }

    public Parser()
    {
      this.ConvertLineBreak = false;

      InlineNotationHandlers.Add("inline plugin", new InlineNotationHandler {
        // &name(args);
        Regex = new Regex(inlinePluginRegexString + ";"),
        Parser = delegate(Match match) {
          var name = match.Groups["name"].Value;
          var args = match.Groups["args"].Value;

          if (0 < args.Length) {
            return new InlinePlugin(name, Csv.Split(args), new Node[] {});
          }
          else {
            if (inlinePluginNames.Contains(name))
              return new InlinePlugin(name, new string[] {});
            else
              // entity (&amp; etc..)
              return new EntityReference(name);
          }
        }
      });

      InlineNotationHandlers.Add("mail address", new InlineNotationHandler {
        // TODO: more strict
        Regex = new Regex(@"\w[a-zA-Z0-9_\-\.]+@[a-zA-Z0-9_\-\.]+\.[a-zA-Z0-9_\-\.]+\w"),
        Parser = delegate(Match match) { return new Anchor(string.Format("mailto:{0}", match.Value), new[] {new Text(match.Value)}); },
      });

      InlineNotationHandlers.Add("page link", new InlineNotationHandler {
        // [[title:href]]
        Regex = new Regex(@"\[{2}(?!\[)(?<title>.+?):(?<href>s?https?://(?:(?!\]\]).)+)\]{2}"),
        Parser = delegate(Match match) {
          return new Anchor(match.Groups["href"].Value,
                            ParseInline(match.Groups["title"].Value));
        }
      });

      InlineNotationHandlers.Add("alias", new InlineNotationHandler {
        // [[name>aliasto]]
        Regex = new Regex(@"\[{2}(?!\[)(?<name>[^\>\]]+)(\>(?<aliasto>[^\]]+))?\]{2}"),
        Parser = delegate(Match match) {
          if (match.Groups["aliasto"].Success)
            return new Alias(match.Groups["name"].Value,
                             match.Groups["aliasto"].Value);
          else
            return new Alias(match.Groups["name"].Value);
        }
      });
    }

    protected Container<Node> GetLastBlockContainable(Document document)
    {
      var blockContainable = (Container<Node>)document;

      for (;;) {
        if (blockContainable.LastChild is DefinitionList)
          blockContainable = (blockContainable.LastChild as DefinitionList).LastChild as DefinitionListItem;
        else if (blockContainable.LastChild is DefinitionListItem)
          blockContainable = blockContainable.LastChild as DefinitionListItem;
        else if (blockContainable.LastChild is OrderedList)
          blockContainable = (blockContainable.LastChild as OrderedList).LastChild as ListItem;
        else if (blockContainable.LastChild is UnorderedList)
          blockContainable = (blockContainable.LastChild as UnorderedList).LastChild as ListItem;
        else
          break;
      }

      return blockContainable;
    }

    protected override void PreParseLine(string line, Document document)
    {
      var multilineBlockPlugin = GetLastBlockContainable(document).LastChild as MultilineBlockPlugin;

      if (multilineBlockPlugin != null) {
        if (line == "}}") {
          // end of multiline plugin
          document.Nodes.Add(new EmptyLine());
        }
        else {
          if (multilineBlockPlugin.IsLastArgumentEmpty)
            multilineBlockPlugin.Append(line);
          else
            multilineBlockPlugin.Append(Environment.NewLine, line);
        }
      }
      else {
        base.PreParseLine(line, document);
      }
    }

    protected override bool TryParseNestableBlock(string line, Document document)
    {
      switch (line[0]) {
        case '-':
        case '+': {
          var list = GetLastBlockContainable(document) as ListItem;

          if (list == null) {
            return base.TryParseNestableBlock(line, document);
          }
          else {
            if (line[0] == '-') {
              var level = 0;

              if (line == "-")
                level = 1;
              else if (line == "--")
                level = 2;
              else if (line == "---")
                level = 3;
              else
                return base.TryParseNestableBlock(line, document);

              ParseEnumListBlockPlaceHolder<UnorderedList>(level, list.Parent as UnorderedList);

              return true;
            }
            else {
              var level = 0;

              if (line == "+")
                level = 1;
              else if (line == "++")
                level = 2;
              else if (line == "+++")
                level = 3;
              else
                return base.TryParseNestableBlock(line, document);

              ParseEnumListBlockPlaceHolder<OrderedList>(level, list.Parent as OrderedList);

              return true;
            }
          }
        }

        case ':':
          if (line.IndexOf('|', 1) < 0)
            return false;

          ParseDefinitionListBlock(line, document);
          return true;

        case '>':
          ParseQuoteBlock(line, document);
          return true;

        default:
          return base.TryParseNestableBlock(line, document);
      }
    }

    // FIXME:
    private void ParseEnumListBlockPlaceHolder<TList>(int level, TList parent) where TList : ListItem, new()
    {
      for (;0 < level; level--) {
        var li = new ListItem();

        parent.Nodes.Add(li);

        if (1 < level) {
          var el = new TList();

          li.Nodes.Add(el);

          parent = el;
        }
      }
    }

    protected override Node ParseLine(string line, Document document)
    {
      if (line.Length == 0)
        return ParseEmptyLine();

      Node parsed = null;

      // TODO: ブロック要素の入れ子の考慮
      switch (line[0]) {
        case ' ':
          ParsePreformattedBlock(line, document);
          return null;
        case '~':
          ParseParagraphBlock(line, document);
          return null;
        case '|':
          ParseTableBlock(line, document);
          return null;
        case ',':
          parsed = ParseCsvTableBlock(line);
          break;
        case '#':
          if (blockPluginRegex.IsMatch(line))
            ParsePluginBlock(line, document);
          else
            ParseParagraphBlock(line, document);
          return null;
        default:
          if (line.StartsWith("//")) {
            parsed = ParseCommentBlock(line);
          }
          else if (line.StartsWith("----")) {
            parsed = new HorizontalRule();
          }
          else if (line.StartsWith("LEFT:")) {
            parsed = new AlignedParagraph("left", ParseInline(line.Substring(5)));
          }
          else if (line.StartsWith("RIGHT:")) {
            parsed = new AlignedParagraph("right", ParseInline(line.Substring(6)));
          }
          else if (line.StartsWith("CENTER:")) {
            parsed = new AlignedParagraph("center", ParseInline(line.Substring(7)));
          }
          else {
            ParseParagraphBlock(line, document);
            return null;
          }
          break;
      }

      if (parsed == null)
        return base.ParseLine(line, document);
      else
        return parsed;
    }

    protected override Container<Node> ParseHeaderBlock(string line)
    {
      var match = headerAnchorRegex.Match(line);
      HeaderAnchor anchor = null;

      if (match.Success) {
        anchor = new HeaderAnchor(match.Groups["id"].Value);

        line = line.Substring(0, match.Index);
      }

      var header = base.ParseHeaderBlock(line);

      if (anchor != null)
        header.Nodes.Add(anchor);

      return header;
    }

    protected override bool TryParseUnorderedListBlock(string line, Document document)
    {
      if (line.StartsWith("----"))
        return false;
      else
        return base.TryParseUnorderedListBlock(line, document);
    }

    private void ParseDefinitionListBlock(string line, Container<Node> parent)
    {
      var list = parent.LastChild as DefinitionList;

      if (list == null) {
        list = new DefinitionList();
        parent.Nodes.Add(list);
      }

      line = line.Substring(1);

      if (0 < line.Length && line[0] == ':') {
        ParseDefinitionListBlock(line, list);
      }
      else {
        // 先にインラインの記法を処理してから分割位置を判断する
        var termNodes = new List<Node>();
        var descNodes = new List<Node>();
        var container = termNodes;

        foreach (var node in ParseInline(line)) {
          if (node is Text) {
            var text = node as Text;
            var delim = text.Value.IndexOf('|');

            // ノードがTextで、termDelimiterを含む場合は、それをtermとdescriptionの
            // 分割位置と判断する
            if (0 <= delim && !(text is Comment)) {
              container.Add(new Text(text.Value.Substring(0, delim)));

              container = descNodes;

              if (delim + 1 < text.Value.Length)
                container.Add(new Text(text.Value.Substring(delim + 1)));
            }
            else {
              container.Add(node);
            }
          }
          else {
            container.Add(node);
          }
        }

        list.Nodes.Add(new DefinitionListItem(termNodes, descNodes));
      }
    }

    private void ParseQuoteBlock(string line, Container<Node> parent)
    {
      if (line.StartsWith("<"))
        throw new NotImplementedException("end-quoting is not implemented");

      line = line.Substring(1);

      var quot = parent.LastChild as BlockQuotation;

      if (quot == null) {
        quot = new BlockQuotation();
        parent.Nodes.Add(quot);
      }

      if (0 < line.Length && line[0] == '>')
        ParseQuoteBlock(line, quot);
      else
        quot.Nodes.AddRange(ParseInline(line));
    }

    private Node ParsePreformattedBlock(string line, Document document)
    {
      var container = GetLastBlockContainable(document);
      var pre = container.LastChild as Preformatted;

      if (pre == null) {
        pre = new Preformatted(line.Substring(1), true);
        container.Nodes.Add(pre);
      }
      else {
        pre.Append(Environment.NewLine, line.Substring(1));
      }

      return pre;
    }

    private Node ParseEmptyLine()
    {
      return new EmptyLine();
    }

    private Container<Node> ParseParagraphBlock(string line, Document document)
    {
      var container = GetLastBlockContainable(document);
      var nodes = line.StartsWith("~")
        ? ParseInline(line.Substring(1))
        : ParseInline(line);

      var p = container.LastChild as Paragraph;

      if (p == null) {
        if (container == document) {
          p = new Paragraph(nodes);
          document.Nodes.Add(p);
        }
        else {
          if (ConvertLineBreak && 0 < container.Nodes.Count)
            container.Nodes.Add(new ForcedLineBreak());

          container.Nodes.AddRange(nodes);
        }
      }
      else {
        p.Nodes.AddRange(nodes);
      }

      return p;
    }

    protected virtual Node ParseTableBlock(string line, Document document)
    {
      if (line.IndexOf("|", 1) < 0)
        return ParseParagraphBlock(line, document);

      line = line.Substring(1);

      var container = GetLastBlockContainable(document);

      // table
      var table = container.LastChild as Table;

      if (table == null) {
        table = new Table();
        container.Nodes.Add(table);
      }

      // row
      TableRow row = null;

      if (line.EndsWith("|h")) {
        line = line.Substring(0, line.Length - 2);
        row = new TableHeaderRow();
      }
      else if (line.EndsWith("|f")) {
        line = line.Substring(0, line.Length - 2);
        row = new TableFooterRow();
      }
      else {
        if (line.EndsWith("|"))
          line = line.Substring(0, line.Length - 1);
        row = new TableRow();
      }

      table.Nodes.Add(row);

      // columns
      foreach (var column in line.Split(new char[] {'|'})) {
        if (column == "~")
          row.Nodes.Add(new TableColumn(new Node[] {}, 0, -1));
        else if (column == ">")
          row.Nodes.Add(new TableColumn(new Node[] {}, 1, 0));
        else if (column.StartsWith("~"))
          row.Nodes.Add(new TableHeaderColumn(ParseInline(column.Substring(1))));
        else
          row.Nodes.Add(new TableColumn(ParseInline(column)));
      }

      return row;
    }

    private Node ParseCsvTableBlock(string line)
    {
      throw new NotImplementedException("csv table is not implemented");
    }

    private Node ParsePluginBlock(string line, Document document)
    {
      var container = GetLastBlockContainable(document);
      var multiline = line.EndsWith("{{");
      var argsStarts = line.IndexOf("(");
      Plugin plugin;

      if (0 < argsStarts) {
        var argsEnds = multiline ? line.LastIndexOf(")") : line.IndexOf(")", argsStarts + 1);
        var name = line.Substring(1, argsStarts - 1);
        var args = (0 <= argsEnds)
          ? line.Substring(argsStarts + 1, argsEnds - argsStarts - 1)
          : line.Substring(argsStarts);

        if (multiline)
          plugin = new MultilineBlockPlugin(name, Csv.Split(args));
        else
          plugin = new BlockPlugin(name, Csv.Split(args));
      }
      else {
        if (multiline)
          plugin = new MultilineBlockPlugin(line.Substring(1, line.Length - 3), new string[] {});
        else
          plugin = new BlockPlugin(line.Substring(1), new string[] {});
      }

      container.Nodes.Add(plugin);

      return plugin;
    }

    private Node ParseCommentBlock(string line)
    {
      return new Comment(line.Substring(2));
    }

    protected override IEnumerable<Node> ParseInline(string line)
    {
      return ParseBrackets((IList<Node>)base.ParseInline(line)); // XXX
    }

    private delegate Node BracketParser(IEnumerable<Node> nodes, Match openBracketMatch, Match closeBracketMatch);

    private class BracketHanlder {
      public Regex OpenBracketRegex = null;
      public Regex CloseBracketRegex = null;
      public BracketParser Parser;
      public bool IsSymmetricBracket = false;
    }

    private bool bracketHandlersInitialzed = false;
    private BracketHanlder[] bracketHandlers = new[] {
      new BracketHanlder() { // annotation
        OpenBracketRegex = new Regex(@"\({2}"),
        CloseBracketRegex   = new Regex(@"\){2}"),
        Parser = delegate(IEnumerable<Node> nodes, Match openBracketMatch, Match closeBracketMatch) {
          return new Annotation(nodes);
        }
      },
      new BracketHanlder() { // inline plugin with text
        OpenBracketRegex = new Regex(inlinePluginRegexString + "{"),
        CloseBracketRegex   = new Regex(@"};"),
        Parser = delegate(IEnumerable<Node> nodes, Match openBracketMatch, Match closeBracketMatch) {
          return new InlinePlugin(openBracketMatch.Groups["name"].Value, Csv.Split(openBracketMatch.Groups["args"].Value), nodes);
        }
      },
      new BracketHanlder() { // COLOR(){}
        OpenBracketRegex = new Regex(@"COLOR\((?<arg>#\d+)\){"),
        CloseBracketRegex   = new Regex(@"}"),
        Parser = delegate(IEnumerable<Node> nodes, Match openBracketMatch, Match closeBracketMatch) {
          return new InlinePlugin("color", new[] {openBracketMatch.Groups["arg"].Value}, nodes);
        }
      },
      new BracketHanlder() { // SIZE(){}
        OpenBracketRegex = new Regex(@"SIZE\((?<arg>\d+)\){"),
        CloseBracketRegex   = new Regex(@"}"),
        Parser = delegate(IEnumerable<Node> nodes, Match openBracketMatch, Match closeBracketMatch) {
          return new InlinePlugin("size", new[] {openBracketMatch.Groups["arg"].Value}, nodes);
        }
      },
      new BracketHanlder() { // inserted text (underline)
        OpenBracketRegex = new Regex(@"\%{3}"),
        Parser = delegate(IEnumerable<Node> nodes, Match openBracketMatch, Match closeBracketMatch) {
          return new InsertedText(nodes);
        }
      },
      new BracketHanlder() { // deleted text (strike through)
        OpenBracketRegex = new Regex(@"\%{2}"),
        Parser = delegate(IEnumerable<Node> nodes, Match openBracketMatch, Match closeBracketMatch) {
          return new DeletedText(nodes);
        }
      },
      new BracketHanlder() { // emphasis (italic)
        OpenBracketRegex = new Regex(@"\'{3}"),
        Parser = delegate(IEnumerable<Node> nodes, Match openBracketMatch, Match closeBracketMatch) {
          return new Emphasis(nodes);
        }
      },
      new BracketHanlder() { // strong emphasis (bold)
        OpenBracketRegex = new Regex(@"\'{2}"),
        Parser = delegate(IEnumerable<Node> nodes, Match openBracketMatch, Match closeBracketMatch) {
          return new StrongEmphasis(nodes);
        }
      },
    };

    private class BracketParsingContext {
      public BracketHanlder Handler;
      public Match OpenBracketMatch;
      public int OpenBracketNodeIndex;
      public Match CloseBracketMatch;
      public int CloseBracketNodeIndex;

      public BracketParsingContext(BracketHanlder handler, Match openBracketMatch, int openBracketNodeIndex)
      {
        this.Handler = handler;
        this.OpenBracketMatch = openBracketMatch;
        this.OpenBracketNodeIndex = openBracketNodeIndex;
      }
    }

    private IEnumerable<Node> ParseBrackets(IList<Node> nodes)
    {
      if (!bracketHandlersInitialzed) {
        foreach (var handler in bracketHandlers) {
          if (handler.CloseBracketRegex == null) {
            handler.IsSymmetricBracket = true;
            handler.CloseBracketRegex = handler.OpenBracketRegex;
          }
        }

        bracketHandlersInitialzed = true;
      }

      for (;;) {
        // search innermost bracket
        var innermost = GetInnermostBracket(nodes);

        if (innermost == null)
          return nodes; // nothing matched(no brackets)

        nodes = ParseBracket(nodes, innermost);

        // reparse
      }
    }

    private List<BracketParsingContext> GetInnerBrackets(IList<Node> nodes)
    {
      var openedBracketContexts = new List<BracketParsingContext>();

      foreach (var handler in bracketHandlers) {
        for (var i = 0; i < nodes.Count; i++) {
          var text = nodes[i] as Text;

          if (text == null)
            continue;

          if (handler.IsSymmetricBracket) {
            var firstMatched = handler.OpenBracketRegex.Match(text.Value);

            if (firstMatched.Success) {
              openedBracketContexts.Add(new BracketParsingContext(handler, firstMatched, i));
              break;
            }
          }
          else {
            // TODO: RegexOptions.RightToLeft
            var matches = handler.OpenBracketRegex.Matches(text.Value);

            if (0 < matches.Count) {
              openedBracketContexts.Add(new BracketParsingContext(handler, matches[matches.Count - 1], i));
              break;
            }
          }
        }
      }

      if (openedBracketContexts.Count == 0)
        return null; // no brackets

      var closedBracketContexts = new List<BracketParsingContext>();

      foreach (var opened in openedBracketContexts) {
        for (var i = opened.OpenBracketNodeIndex; i < nodes.Count; i++) {
          var text = nodes[i] as Text;

          if (text == null)
            continue;

          var match = (i == opened.OpenBracketNodeIndex)
            ? opened.Handler.CloseBracketRegex.Match(text.Value, opened.OpenBracketMatch.Index + opened.OpenBracketMatch.Length)
            : opened.Handler.CloseBracketRegex.Match(text.Value);

          if (match.Success) {
            opened.CloseBracketMatch = match;
            opened.CloseBracketNodeIndex = i;

            closedBracketContexts.Add(opened);

            break;
          }
        }
      }

      if (closedBracketContexts.Count == 0)
        return null; // no brackets (probably not closed)
      else
        return closedBracketContexts;
    }

    private BracketParsingContext GetInnermostBracket(IList<Node> nodes)
    {
      var bracketContexts = GetInnerBrackets(nodes);

      if (bracketContexts == null)
        return null;
      else if (bracketContexts.Count == 1)
        return bracketContexts[0];

      // find innermost (=shortest) bracket
      bracketContexts.Sort(delegate(BracketParsingContext x, BracketParsingContext y) {
        if (x == y)
          return 0;

        var openComparison = x.OpenBracketNodeIndex - y.OpenBracketNodeIndex;

        if (openComparison == 0)
          openComparison = x.OpenBracketMatch.Index - y.OpenBracketMatch.Index;

        var closeComparison = x.CloseBracketNodeIndex - y.CloseBracketNodeIndex;

        if (closeComparison == 0)
          closeComparison = (x.CloseBracketMatch.Index + x.CloseBracketMatch.Length) - (y.CloseBracketMatch.Index + y.CloseBracketMatch.Length);

        if (openComparison < 0 && 0 < closeComparison)
          return 1;
        else
          return -1;
      });

      return bracketContexts[0];
    }

    private IList<Node> ParseBracket(IList<Node> nodes, BracketParsingContext context)
    {
      var leadingNodes = new List<Node>();
      var parsingNodes = new List<Node>();
      var trailingNodes = new List<Node>();

      for (var i = 0; i < context.OpenBracketNodeIndex; i++) {
        leadingNodes.Add(nodes[i]);
      }

      var startText = nodes[context.OpenBracketNodeIndex] as Text;

      if (0 < context.OpenBracketMatch.Index)
        leadingNodes.Add(new Text(startText.Value.Substring(0, context.OpenBracketMatch.Index)));

      {
        var start = context.OpenBracketMatch.Index + context.OpenBracketMatch.Length;

        if (context.OpenBracketNodeIndex == context.CloseBracketNodeIndex)
          parsingNodes.Add(new Text(startText.Value.Substring(start, context.CloseBracketMatch.Index - start)));
        else
          parsingNodes.Add(new Text(startText.Value.Substring(start)));
      }

      for (var i = context.OpenBracketNodeIndex + 1; i < context.CloseBracketNodeIndex; i++) {
        parsingNodes.Add(nodes[i]);
      }

      var endText = nodes[context.CloseBracketNodeIndex] as Text;

      {
        var end = context.CloseBracketMatch.Index + context.CloseBracketMatch.Length;

        if (context.OpenBracketNodeIndex != context.CloseBracketNodeIndex && 0 < context.CloseBracketMatch.Index)
          parsingNodes.Add(new Text(endText.Value.Substring(0, context.CloseBracketMatch.Index)));

        if (end < endText.Value.Length)
          trailingNodes.Add(new Text(endText.Value.Substring(end)));
      }

      for (var i = context.CloseBracketNodeIndex + 1; i < nodes.Count; i++) {
        trailingNodes.Add(nodes[i]);
      }

      var ret = new List<Node>();

      ret.AddRange(leadingNodes);
      ret.Add(context.Handler.Parser(parsingNodes, context.OpenBracketMatch, context.CloseBracketMatch));
      ret.AddRange(trailingNodes);

      return ret;
    }
  }
}
