// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Xml;

using Smdn.Formats.Notations;
using Smdn.Formats.Notations.Dom;
using Smdn.Formats.Notations.PukiWiki;
using Smdn.Formats.Notations.PukiWiki.Dom;

namespace Smdn.Formats.Notations.PukiWiki {
  public class XhtmlFormatter : Notations.XhtmlFormatter {
    public const string PukiWikiExtensionNamespace = "http://smdn.invisiblefulmoon.net/works/tools/pakina/notations#pukiwiki"; // XXX

    protected override void RegisterPrefixes()
    {
      base.RegisterPrefixes();
      base.NamespaceManager.AddNamespace("pukiwiki", PukiWikiExtensionNamespace);
    }

    protected override IEnumerable<XmlNode> FormatNode(XmlDocument xml, Node node)
    {
      if (node is Plugin)
        return FormatPluginNode(xml, node as Plugin);
      else if (node is AlignedParagraph)
        return FormatAlignedParagraphNode(xml, node as AlignedParagraph);
      else if (node is Alias)
        return FormatAliasNode(xml, node as Alias);
      else if (node is HeaderAnchor)
        return FormatHeaderAnchor(xml, node as HeaderAnchor);
      else
        return base.FormatNode(xml, node);
    }

    private readonly string[] refParams = new[] {
      "nolink",
      "left",
      "center",
      "right",
      "around",
      "wrap",
      "nowrap",
    };

    protected virtual IEnumerable<XmlNode> FormatPluginNode(XmlDocument xml, Plugin plugin)
    {
      switch (plugin.Name) {
        case "t":
          return new XmlNode[] {xml.CreateTextNode("\t")};

        case "br":
          return new XmlNode[] {xml.CreateElement("br", XhtmlNamespaceUri)};

        case "heart":
          return new XmlNode[] {xml.CreateEntityReference("hearts")};

        case "new": {
          return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("span", XhtmlNamespaceUri), FormatNodes(xml, plugin.Nodes))};
        }

        // face mark
        case "smile":
        case "bigsmile":
        case "huh":
        case "oh":
        case "wink":
        case "sad":
        case "worried": {
          var node = xml.CreateElement("span", XhtmlNamespaceUri);

          switch (plugin.Name) {
            case "oh": case "sad": case "worried":
              // 'WHITE FROWNING FACE' (U+2639)
              node.AppendChild(xml.CreateTextNode(((char)0x2639).ToString()));
              break;
            case "huh": case "wink":
              // 'BLACK SMILING FACE' (U+263B)
              node.AppendChild(xml.CreateTextNode(((char)0x263b).ToString()));
              break;
            default:
              // 'WHITE SMILING FACE' (U+263A)
              node.AppendChild(xml.CreateTextNode(((char)0x263a).ToString()));
              break;
          }

          var sup = node.AppendChild(xml.CreateElement("sup", XhtmlNamespaceUri));

          sup.AppendChild(xml.CreateTextNode(plugin.Name));

          return new XmlNode[] {node};
        }

        case "sub": {
          return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("sub", XhtmlNamespaceUri), FormatNodes(xml, plugin.Nodes))};
        }

        case "sup": {
          return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("sup", XhtmlNamespaceUri), FormatNodes(xml, plugin.Nodes))};
        }

        case "ruby": {
          var ruby = xml.CreateElement("ruby", XhtmlNamespaceUri);

          ruby.AppendChild(XmlNodeExtension.AppendChildren(xml.CreateElement("rb", XhtmlNamespaceUri), FormatNodes(xml, plugin.Nodes)));
          ruby.AppendChild(XmlNodeExtension.AppendChildren(xml.CreateElement("rp", XhtmlNamespaceUri), new[] {xml.CreateTextNode("(")}));
          ruby.AppendChild(XmlNodeExtension.AppendChildren(xml.CreateElement("rt", XhtmlNamespaceUri), new[] {xml.CreateTextNode(plugin.Arguments[0])}));
          ruby.AppendChild(XmlNodeExtension.AppendChildren(xml.CreateElement("rp", XhtmlNamespaceUri), new[] {xml.CreateTextNode(")")}));

          return new XmlNode[] {ruby};
        }

        case "ref":
        {
          string filename = null;
          string title = null;
          string align = null;

          foreach (var arg in plugin.Arguments) {
            var handled = false;

            foreach (var refparam in refParams) {
              if (arg != refparam)
                continue;

              switch (refparam) {
                case "left":
                case "center":
                case "right":
                  align = refparam;
                  break;
                default:
                  Console.Error.WriteLine("ref parameter '{0}' is not formatted", refparam);
                  break;
              }

              handled = true;
            }

            if (handled)
              continue;

            if (filename == null) {
              filename = arg;
              continue;
            }

            if (title == null) {
              title = arg;
              continue;
            }
          }

          XmlNode node = null;

          switch (Path.GetExtension(filename).ToLower()) {
            case ".jpg":
            case ".png":
            case ".gif":
            case ".tif":
            case ".bmp":
              node = xml.CreateElement("img", XhtmlNamespaceUri);
              node.Attributes.Append(xml.CreateAttribute("src")).Value = filename;
              break;
            default:
              node = xml.CreateElement("a", XhtmlNamespaceUri);
              node.Attributes.Append(xml.CreateAttribute("href")).Value = filename;
              if (title == null)
                node.AppendChild(xml.CreateTextNode(filename));
              else
                node.AppendChild(xml.CreateTextNode(title));
              break;
          }

          if (title == null) {
            if (node.Name == "img")
              node.Attributes.Append(xml.CreateAttribute("alt")).Value = filename;
            node.Attributes.Append(xml.CreateAttribute("title")).Value = filename;
          }
          else {
            if (node.Name == "img")
              node.Attributes.Append(xml.CreateAttribute("alt")).Value = title;
            node.Attributes.Append(xml.CreateAttribute("title")).Value = title;
          }

          if (plugin is InlinePlugin) {
            return new XmlNode[] {node};
          }
          else {
            var parentNode = xml.CreateElement("div", XhtmlNamespaceUri);

            if (!string.IsNullOrEmpty(align))
              parentNode.Attributes.Append(xml.CreateAttribute("style")).Value = string.Format("text-align: {0};", align);

            parentNode.AppendChild(node);

            return new XmlNode[] {parentNode};
          }
        }

        case "color":
        {
          XmlNode node;

          if (plugin is InlinePlugin)
            node = xml.CreateElement("span", XhtmlNamespaceUri);
          else
            node = xml.CreateElement("div", XhtmlNamespaceUri);

          XmlNodeExtension.AppendChildren(node, FormatNodes(xml, plugin.Nodes));

          string style;

          if (plugin.Arguments.Length == 1)
            style = string.Format("color: {0};", plugin.Arguments[0]);
          else if (plugin.Arguments.Length == 2)
            style = string.Format("color: {0}; background-color: {1};", plugin.Arguments[0], plugin.Arguments[1]);
          else
            style = null;

          if (style != null)
            node.Attributes.Append(xml.CreateAttribute("style")).Value = style;

          return new XmlNode[] {node};
        }

        case "size": {
          XmlNode node;

          if (plugin is InlinePlugin)
            node = xml.CreateElement("span", XhtmlNamespaceUri);
          else
            node = xml.CreateElement("div", XhtmlNamespaceUri);

          XmlNodeExtension.AppendChildren(node, FormatNodes(xml, plugin.Nodes));

          node.Attributes.Append(xml.CreateAttribute("style")).Value = string.Format("font-size: {0}px;", plugin.Arguments[0]);

          return new XmlNode[] {node};
        }

        default:
        {
          var unformatted = xml.CreateElement("pukiwiki:plugin", PukiWikiExtensionNamespace);

          unformatted.Attributes.Append(xml.CreateAttribute("pukiwiki:type", PukiWikiExtensionNamespace)).Value = (plugin is InlinePlugin) ? "inline" : "block";
          unformatted.Attributes.Append(xml.CreateAttribute("pukiwiki:name", PukiWikiExtensionNamespace)).Value = plugin.Name;
          unformatted.Attributes.Append(xml.CreateAttribute("pukiwiki:arguments", PukiWikiExtensionNamespace)).Value = Csv.Join(plugin.Arguments);

          if (plugin is BlockPlugin) {
            unformatted.AppendChild(xml.CreateTextNode(string.Format("#{0}", plugin.Name)));

            if (plugin is MultilineBlockPlugin) {
              unformatted.AppendChild(xml.CreateTextNode("{{"));
              unformatted.AppendChild(xml.CreateCDataSection((plugin as MultilineBlockPlugin).LastArgument));
              unformatted.AppendChild(xml.CreateTextNode("}}"));
            }
          }
          else {
            if (0 < plugin.Nodes.Count)
              XmlNodeExtension.AppendChildren(unformatted, FormatNodes(xml, plugin.Nodes));
            else
              unformatted.AppendChild(xml.CreateTextNode(string.Format("&{0};", plugin.Name)));
          }

          return new[] {unformatted};
        }
      }
    }

    private IEnumerable<XmlNode> FormatAlignedParagraphNode(XmlDocument xml, AlignedParagraph p)
    {
      var node = XmlNodeExtension.AppendChildren(xml.CreateElement("div", XhtmlNamespaceUri), FormatNodes(xml, p.Nodes));

      node.Attributes.Append(xml.CreateAttribute("style")).Value = string.Format("text-align: {0};", p.Alignment);

      return new XmlNode[] {node};
    }

    protected virtual IEnumerable<XmlNode> FormatAliasNode(XmlDocument xml, Alias alias)
    {
      var unformatted = xml.CreateElement("pukiwiki:alias", PukiWikiExtensionNamespace);

      unformatted.Attributes.Append(xml.CreateAttribute("pukiwiki:name", PukiWikiExtensionNamespace)).Value = alias.Name;
      unformatted.Attributes.Append(xml.CreateAttribute("pukiwiki:aliasto", PukiWikiExtensionNamespace)).Value = alias.AliasTo;

      if (string.IsNullOrEmpty(alias.AliasTo))
        unformatted.AppendChild(xml.CreateTextNode(string.Format("[[{0}]]", alias.Name)));
      else
        unformatted.AppendChild(xml.CreateTextNode(string.Format("[[{0}>{1}]]", alias.Name, alias.AliasTo)));

      return new XmlNode[] {unformatted};
    }

    private IEnumerable<XmlNode> FormatHeaderAnchor(XmlDocument xml, HeaderAnchor anchor)
    {
      var a = XmlNodeExtension.FindFirstElement("a", XhtmlNamespaceUri, base.FormatAnchorNode(xml, anchor));

      a.AppendChild(xml.CreateEntityReference("dagger"));

      a.Attributes.Append(xml.CreateAttribute("class")).Value = "anchor_super";

      return new XmlNode[] {a};
    }
  }
}
