// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;

using Smdn.Formats.Notations.Dom;

namespace Smdn.Formats.Notations {
  public abstract class XhtmlFormatter : IFormatter<XmlDocument> {
    public const string XhtmlNamespaceUri = "http://www.w3.org/1999/xhtml";

    protected XmlNamespaceManager NamespaceManager {
      get { return nsmgr; }
    }

    private XmlNamespaceManager nsmgr = null;

    public XmlDocument Format(Document document)
    {
      return Format(document, "body", string.Empty, XhtmlNamespaceUri);
    }

    public XmlDocument Format(Document document, string documentElement, string prefixDocumentElement, string nsUriDocumentElement)
    {
      var xml = new XmlDocument();

      try {
        nsmgr = new XmlNamespaceManager(xml.NameTable);
        nsmgr.PushScope();

        if (!string.IsNullOrEmpty(prefixDocumentElement) && string.IsNullOrEmpty(nsUriDocumentElement))
          nsmgr.AddNamespace(prefixDocumentElement, nsUriDocumentElement);

        RegisterPrefixes();

        var body = xml.AppendChild(xml.CreateElement(documentElement, nsUriDocumentElement));
        var namespaces = new Dictionary<string, XmlAttribute>();

        foreach (string ns in nsmgr) {
          if (string.Empty.Equals(ns) || ns == "xml") {
            continue;
          }
          else if (ns == "xmlns") {
            body.Attributes.Append(xml.CreateAttribute("xmlns")).Value = XhtmlNamespaceUri;
          }
          else {
            var attr = body.Attributes.Append(xml.CreateAttribute("xmlns:" + ns));

            attr.Value = nsmgr.LookupNamespace(ns);

            namespaces.Add(attr.Value, attr);
          }
        }

        foreach (var node in document.Nodes) {
          XmlNodeExtension.AppendChildren(body, FormatNode(xml, node));
        }

        foreach (var ns in namespaces) {
          if (body.SelectSingleNode(string.Format("//*[namespace-uri()='{0}']", ns.Key)) == null)
            body.Attributes.Remove(ns.Value);
        }

        return xml;
      }
      finally {
        nsmgr.PopScope();
        nsmgr = null;
      }
    }

    protected virtual void RegisterPrefixes()
    {
      nsmgr.AddNamespace(string.Empty, XhtmlNamespaceUri);
    }

    protected virtual IEnumerable<XmlNode> FormatNode(XmlDocument xml, Node node)
    {
      if (node is Text) {
        return FormatTextNode(xml, node as Text);
      }
      else if (node is Preformatted) {
        return FormatPreformattedNode(xml, node as Preformatted);
      }
      else if (node is Table) {
        return new XmlNode[] {FormatTableNode(xml, node as Table)};
      }
      else if (node is Container<Node>) {
        var container = node as Container<Node>;
        if (container is ListItem)
          return FormatListNode(xml, container as ListItem);
        else if (node is DefinitionListItem)
          return FormatDefinitionListNode(xml, node as DefinitionListItem);
        else if (container is Header)
          return FormatHeaderNode(xml, container as Header);
        else if (container is BlockQuotation)
          return FormatBlockQuotationNode(xml, container as BlockQuotation);
        else if (container is Paragraph)
          return FormatParagraphNode(xml, container as Paragraph);
        else if (container is Anchor)
          return FormatAnchorNode(xml, container as Anchor);
        else if (container is Emphasis)
          return FormatEmphasisNode(xml, container as Emphasis);
        else if (container is InsertedText)
          return FormatInsertedTextNode(xml, container as InsertedText);
        else if (container is DeletedText)
          return FormatDeletedTextNode(xml, container as DeletedText);
        else if (container is Inline)
          return FormatNodes(xml, container.Nodes);
        else if (node is Annotation)
          return FormatAnnotationNode(xml, node as Annotation);
        else
          return ProcessUnformattedNode(xml, container);
      }
      else if (node is HorizontalRule) {
        return new XmlNode[] {xml.CreateElement("hr", XhtmlNamespaceUri)};
      }
      else if (node is ForcedLineBreak) {
        return new XmlNode[] {xml.CreateElement("br", XhtmlNamespaceUri)};
      }
      else if (node is EmptyLine) {
        return new XmlNode[] {};
      }
      else if (node is EntityReference) {
        return FormatEntityReferenceNode(xml, node as EntityReference);
      }
      else if (node is InlineFrame) {
        return FormatInlineFrameNode(xml, node as InlineFrame);
      }
      else if (node is Image) {
        return FormatImageNode(xml, node as Image);
      }
      else {
        return ProcessUnformattedNode(xml, node);
      }
    }

    protected virtual IEnumerable<XmlNode> ProcessUnformattedNode(XmlDocument xml, Node node)
    {
      return new XmlNode[] {xml.CreateComment(node.ToString())};
    }

    protected virtual IEnumerable<XmlNode> FormatNodes(XmlDocument xml, IEnumerable<Node> nodes)
    {
      var xmlNodes = new List<XmlNode>();

      foreach (var node in nodes) {
        xmlNodes.AddRange(FormatNode(xml, node));
      }

      return xmlNodes;
    }

    private IEnumerable<XmlNode> FormatTextNode(XmlDocument xml, Text text)
    {
      if (text is Comment)
        return new XmlNode[] {xml.CreateComment(text.Value)};
      else
        return new XmlNode[] {xml.CreateTextNode(text.Value)};
    }

    protected IEnumerable<XmlNode> FormatPreformattedNode(XmlDocument xml, Preformatted preformatted)
    {
      var pre = xml.CreateElement("pre", XhtmlNamespaceUri);
      var parent = (XmlNode)pre;

      if (preformatted is BlockCode) {
        parent = pre.AppendChild(xml.CreateElement("code", XhtmlNamespaceUri));

        parent.Attributes.Append(xml.CreateAttribute("xml:space")).Value = "preserve";
      }

      //if (preformatted.NeedEscape)
        parent.AppendChild(xml.CreateTextNode(preformatted.ToString()));
      //else
        //pre.AppendChild(xml.ImportNode());
        //throw new NotSupportedException("pre-formatted text block which contains XHTML tags is not supported");

      return new XmlNode[] {pre};
    }

    private IEnumerable<XmlNode> FormatDefinitionListNode(XmlDocument xml, DefinitionListItem dli)
    {
      if (dli is DefinitionList) {
        return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("dl", XhtmlNamespaceUri),
                                                                FormatDefinitionListItemNodes(xml, dli.Nodes))};
      }
      else {
        return ProcessUnformattedNode(xml, dli);
      }
    }

    private IEnumerable<XmlNode> FormatDefinitionListItemNodes(XmlDocument xml, IEnumerable<Node> items)
    {
      var nodes = new List<XmlNode>();

      foreach (var item in items) {
        if (item is DefinitionList) {
          nodes.AddRange(FormatDefinitionListNode(xml, item as DefinitionList));
        }
        else if (item is DefinitionListItem) {
          nodes.Add(XmlNodeExtension.AppendChildren(xml.CreateElement("dt", XhtmlNamespaceUri), FormatNodes(xml, (item as DefinitionListItem).Term)));
          nodes.Add(XmlNodeExtension.AppendChildren(xml.CreateElement("dd", XhtmlNamespaceUri), FormatNodes(xml, (item as DefinitionListItem).Nodes)));
        }
        else {
          nodes.AddRange(ProcessUnformattedNode(xml, item));
        }
      }

      return nodes;
    }

    private IEnumerable<XmlNode> FormatListNode(XmlDocument xml, ListItem li)
    {
      if (li is OrderedList)
        return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("ol", XhtmlNamespaceUri), FormatListItemNodes(xml, li.Nodes))};
      else if (li is UnorderedList)
        return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("ul", XhtmlNamespaceUri), FormatListItemNodes(xml, li.Nodes))};
      else 
        return ProcessUnformattedNode(xml, li);
    }

    private IEnumerable<XmlNode> FormatListItemNodes(XmlDocument xml, IEnumerable<Node> items)
    {
      var nodes = new List<XmlNode>();

      foreach (var item in items) {
        if (item is ListItem)
          nodes.Add(XmlNodeExtension.AppendChildren(xml.CreateElement("li", XhtmlNamespaceUri), FormatNodes(xml, (item as ListItem).Nodes)));
        else
          nodes.AddRange(ProcessUnformattedNode(xml, item));
      }

      return nodes;
    }

    private IEnumerable<XmlNode> FormatHeaderNode(XmlDocument xml, Header h)
    {
      return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement(string.Format("h{0}", h.Level), XhtmlNamespaceUri), FormatNodes(xml, h.Nodes))};
    }

    protected virtual IEnumerable<XmlNode> FormatBlockQuotationNode(XmlDocument xml, BlockQuotation blockquote)
    {
      var bq = xml.CreateElement("blockquote", XhtmlNamespaceUri);

      if (blockquote.Cite != null)
        bq.Attributes.Append(xml.CreateAttribute("cite")).Value = blockquote.Cite.ToString();

      XmlNodeExtension.AppendChildren(bq.AppendChild(xml.CreateElement("div", XhtmlNamespaceUri)), FormatNodes(xml, blockquote.Nodes));

      return new XmlNode[] {bq};
    }

    private IEnumerable<XmlNode> FormatParagraphNode(XmlDocument xml, Paragraph p)
    {
      return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("p", XhtmlNamespaceUri), FormatNodes(xml, p.Nodes))};
    }

    protected IEnumerable<XmlNode> FormatAnchorNode(XmlDocument xml, Anchor a)
    {
      var node = XmlNodeExtension.AppendChildren(xml.CreateElement("a", XhtmlNamespaceUri), FormatNodes(xml, a.Nodes));

      node.Attributes.Append(xml.CreateAttribute("href")).Value = Uri.EscapeUriString(a.Href);

      if (a.Id != null)
        node.Attributes.Append(xml.CreateAttribute("id")).Value = a.Id;
      if (a.Title != null)
        node.Attributes.Append(xml.CreateAttribute("title")).Value = a.Title;

      return new XmlNode[] {node};
    }

    private IEnumerable<XmlNode> FormatEmphasisNode(XmlDocument xml, Emphasis em)
    {
      if (em is StrongEmphasis)
        return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("strong", XhtmlNamespaceUri), FormatNodes(xml, em.Nodes))};
      else
        return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("em", XhtmlNamespaceUri), FormatNodes(xml, em.Nodes))};
    }

    private IEnumerable<XmlNode> FormatInsertedTextNode(XmlDocument xml, InsertedText ins)
    {
      return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("ins", XhtmlNamespaceUri), FormatNodes(xml, ins.Nodes))};
    }

    private IEnumerable<XmlNode> FormatDeletedTextNode(XmlDocument xml, DeletedText del)
    {
      return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("del", XhtmlNamespaceUri), FormatNodes(xml, del.Nodes))};
    }

    private XmlNode FormatTableNode(XmlDocument xml, Table table)
    {
      var tbl = XmlNodeExtension.AppendChildren(xml.CreateElement("table", XhtmlNamespaceUri), FormatTableItemNodes(xml, table.Nodes));

      if (table.Summary != null)
        tbl.Attributes.Append(xml.CreateAttribute("summary")).Value = table.Summary;

      return tbl;
    }

    private IEnumerable<XmlNode> FormatTableItemNodes(XmlDocument xml, IEnumerable<TableItem> items)
    {
      var nodes = new List<XmlNode>();

      // check table format
      TableHeaderRow theadRow = null;
      TableFooterRow tfootRow = null;
      TableCaption caption = null;
      var tbodyRows = new List<TableRow>();

      foreach (var item in items) {
        if (item is TableHeaderRow) {
          if (theadRow != null)
            throw new InvalidOperationException("thead should appear exactly one");
          theadRow = item as TableHeaderRow;
        }
        else if (item is TableFooterRow) {
          if (tfootRow != null)
            throw new InvalidOperationException("tfoot should appear exactly one");
          tfootRow = item as TableFooterRow;
        }
        else if (item is TableCaption) {
          if (caption != null)
            throw new InvalidOperationException("caption should appear exactly one");
          caption = item as TableCaption;
        }
        else if (item is TableRow) {
          tbodyRows.Add(item as TableRow);
        }
        else {
          throw new InvalidOperationException("can't contain non TableItem types");
        }
      }


      // merge colspan
      var theadCols = theadRow == null ? null : FormatTableColspan(theadRow.Nodes.ToArray());
      var tfootCols = tfootRow == null ? null : FormatTableColspan(tfootRow.Nodes.ToArray());

      // caption, thead, tfoot, tbodyの順にノードを追加する
      if (caption != null)
        nodes.Add(XmlNodeExtension.AppendChildren(xml.CreateElement("caption", XhtmlNamespaceUri), FormatNodes(xml, caption.Nodes)));

      if (theadCols != null)
        nodes.Add(XmlNodeExtension.AppendChildren(xml.CreateElement("thead", XhtmlNamespaceUri),
                                                  new [] {XmlNodeExtension.AppendChildren(xml.CreateElement("tr", XhtmlNamespaceUri), FormatTableColumnNodes(xml, theadCols))}));

      if (tfootCols != null)
        nodes.Add(XmlNodeExtension.AppendChildren(xml.CreateElement("tfoot", XhtmlNamespaceUri),
                                                  new [] {XmlNodeExtension.AppendChildren(xml.CreateElement("tr", XhtmlNamespaceUri), FormatTableColumnNodes(xml, tfootCols))}));

      if (0 < tbodyRows.Count) {
        var tbody = xml.CreateElement("tbody", XhtmlNamespaceUri);

        foreach (var row in tbodyRows) {
          var tbodyCols = FormatTableColspan(row.Nodes.ToArray());

          XmlNodeExtension.AppendChildren(tbody,
                         new [] {XmlNodeExtension.AppendChildren(xml.CreateElement("tr", XhtmlNamespaceUri), FormatTableColumnNodes(xml, tbodyCols))});
        }

        nodes.Add(tbody);
      }

      return nodes;
    }

    private TableColumn[] FormatTableColspan(Node[] columns)
    {
      var cols = new TableColumn[columns.Length];

      // convert types
      for (var i = 0; i < columns.Length; i++) {
        if (!(columns[i] is TableColumn))
          throw new InvalidOperationException("can't contain non TableColumn types");
        cols[i] = columns[i] as TableColumn;
      }

      // format
      for (var i = 0; i < cols.Length; i++) {
        if (cols[i] == null)
          continue;

        if (0 < cols[i].ColSpan) {
          var span = cols[i].ColSpan;

          if (span == 1) {
            for (var j = i + 1; j < cols.Length; j++) {
              cols[i].ColSpan++;

              if (cols[j].ColSpan == 1) {
                cols[j] = null;
                continue;
              }
              else {
                cols[j].ColSpan = cols[i].ColSpan;
                cols[i] = cols[j];
                cols[j] = null;
                break;
              }
            }
          }
          else {
            throw new NotImplementedException();
          }
        }
        else if (cols[i].ColSpan < 0) {
          throw new NotImplementedException();
        }
      }

      return cols;
    }

    private IEnumerable<XmlNode> FormatTableColumnNodes(XmlDocument xml, IEnumerable<Node> columns)
    {
      var nodes = new List<XmlNode>();

      foreach (var column in columns) {
        if (column is TableColumn) {
          var col = column as TableColumn;
          var node = XmlNodeExtension.AppendChildren(xml.CreateElement(col is TableHeaderColumn ? "th" : "td", XhtmlNamespaceUri), FormatNodes(xml, col.Nodes));

          if (1 < col.RowSpan)
            node.Attributes.Append(xml.CreateAttribute("rowspan")).Value = col.RowSpan.ToString();
          if (1 < col.ColSpan)
            node.Attributes.Append(xml.CreateAttribute("colspan")).Value = col.ColSpan.ToString();

          nodes.Add(node);
        }
        else if (column != null) {
          nodes.AddRange(ProcessUnformattedNode(xml, column));
        }
      }

      return nodes;
    }

    protected virtual IEnumerable<XmlNode> FormatAnnotationNode(XmlDocument xml, Annotation a)
    {
      return new XmlNode[] {XmlNodeExtension.AppendChildren(xml.CreateElement("em", XhtmlNamespaceUri), FormatNodes(xml, a.Nodes))};
    }

    protected virtual IEnumerable<XmlNode> FormatEntityReferenceNode(XmlDocument xml, EntityReference entity)
    {
      return new XmlNode[] {xml.CreateEntityReference(entity.Name)};
    }

    protected virtual IEnumerable<XmlNode> FormatInlineFrameNode(XmlDocument xml, InlineFrame iframe)
    {
      var node = xml.CreateElement("iframe", XhtmlNamespaceUri);

      node.Attributes.Append(xml.CreateAttribute("src")).Value = Uri.EscapeUriString(iframe.SourceUri.ToString());
      node.Attributes.Append(xml.CreateAttribute("scrolling")).Value = iframe.Scrolling ? "yes" : "no";
      node.Attributes.Append(xml.CreateAttribute("frameborder")).Value = iframe.FrameBorder ? "1" : "0";

      if (iframe.Width != null)
        node.Attributes.Append(xml.CreateAttribute("width")).Value = iframe.Width;
      if (iframe.Height != null)
        node.Attributes.Append(xml.CreateAttribute("height")).Value = iframe.Height;

      foreach (var attr in iframe.Attributes) {
        node.Attributes.Append(xml.CreateAttribute(attr.Key)).Value = attr.Value;
      }

      return new XmlNode[] {node};
    }

    protected virtual IEnumerable<XmlNode> FormatImageNode(XmlDocument xml, Image img)
    {
      var node = xml.CreateElement("img", XhtmlNamespaceUri);

      node.Attributes.Append(xml.CreateAttribute("src")).Value = Uri.EscapeUriString(img.Source);
      node.Attributes.Append(xml.CreateAttribute("alt")).Value = img.AlternativeText;

      if (img.Title != null)
        node.Attributes.Append(xml.CreateAttribute("title")).Value = img.Title;
      if (img.Width != null)
        node.Attributes.Append(xml.CreateAttribute("width")).Value = img.Width;
      if (img.Height != null)
        node.Attributes.Append(xml.CreateAttribute("height")).Value = img.Height;

      return new XmlNode[] {node};
    }
  }
}
